# Author: Oriol Tinto Prims

import sys
import os
from subprocess import call

###################################################################
# Definition of some paths
#
# The toolpath relates to the place where the tool is installed
# This is automatically modified by the Configure script.
# If you want to use this script standalone you must put here the correct path,
# otherwise leave it as it is.
toolpath="/.statelite/tmpfs/gpfs/scratch/bsc32/bsc32870/RemoveLandProcesses"

# Patch to correct the path
toolpath+="/"
# Path to source
srcpath=toolpath+"/src/"
srcpath = os.path.relpath(srcpath, toolpath)
#Path to namelist template and check
templatepath = srcpath+"/template.namelist"
assert os.path.isfile(templatepath), "Namelist template not found in src"
###################################################################


###################################################################
# Comprovation of command line arguments
def Check_Parameters(parameters,resolutions):
   if len(parameters) != 2:
      if using_ece:
         print "Using EC-Earth bahtymetries"
      print "Arguments needed:"
      print "       Resolution ( Options are : %s) " % "".join([res+" " for res in resolutions])
      print "       Number of processors"
      exit(1)
   
   if parameters[0] not in resolutions:
     print "ERROR: %s is not a valid resolution" % parameters[0]
     exit(1)
###################################################################


###################################################################
# Find decomposition in prime numbers
def primes(n):
    primfac = []
    d = 2
    while d*d <= n:
        while (n % d) == 0:
            primfac.append(d)  # supposing you want multiple factors repeated
            n //= d
        d += 1
    if n > 1:
       primfac.append(n)
    return primfac
###################################################################

###################################################################
# Find decomposition of N=a*b as squared as posible
def decompose(n):
   p = primes(n)
   while len(p) > 2:
       t = p[0]
       p = p[1:]
       p[-2] = p[-2]*t
       p.sort()
   p.reverse()
   if len(p) == 2:
      return p
   else:
      return p+[1]
###################################################################
###################################################################
#  Main script
###################################################################
# Using optparse library to get command line options and arguments
from optparse import OptionParser

# Creating the parser
parser = OptionParser()

# Adding options

# Option for generating simplified output
parser.add_option("-s",action="store_true", help="Simplified output",dest="simplified",default=False)
# Option for selecting the length of the list of possible configurations
parser.add_option("-l",type="int",help="Lencght of possible configurations",dest="list_length",default=5)
# Option for using EC-Earth bathymetries instead of NEMO ones.
parser.add_option("-e",help="Use EC-Earth grids",action="store_true",dest="ecearth",default=False)

# Parse arguments and options
(options, args) = parser.parse_args()
# Save the options in adequate variables
simplified = options.simplified
list_length = int(options.list_length)
using_ece = options.ecearth
###################################################################

resolution,numprocs = args

###################################################################

#Grid description for different resolutions for NEMO and EC-Earth grids
if using_ece:
    # grids used in EC-Earth are slightly different
    resolutions = ["ORCA1","ORCA025"]
    grid_jpi =  { 'ORCA1': '362', 'ORCA025': '1442'};
    grid_jpj =  { 'ORCA1': '292', 'ORCA025': '1050'};
    grid_jpk =  { 'ORCA1': '75',  'ORCA025': '75'};
else:
    resolutions = ["ORCA2","ORCA1","ORCA025","ORCA12"]
    grid_jpi =  {'ORCA2': '182', 'ORCA1': '362', 'ORCA025': '1442','ORCA12':'4322'};
    grid_jpj =  {'ORCA2': '149', 'ORCA1': '292', 'ORCA025': '1021','ORCA12':'3059'};
    grid_jpk =  {'ORCA2': '31', 'ORCA1': '75', 'ORCA025': '75','ORCA12':'75'};
###################################################################

# Check parameters
Check_Parameters(args,resolutions)

# Converting to int variable
numprocs = int(numprocs)
# Selecting Bathymetry file acording to resolution and model selected
if using_ece:
    bathymetry_file = toolpath+"/bathymetries/ece_bathy_meter_"+resolution+".nc" 
else:
    bathymetry_file = toolpath+"/bathymetries/nemo_bathy_meter_"+resolution+".nc" 
    
# Puting grid description acording to selected resolution and model.
bathymetry_file = os.path.relpath(bathymetry_file, toolpath)
jpi = grid_jpi[resolution]
jpj = grid_jpj[resolution]
jpk = grid_jpk[resolution]

# Changing variables in template
with open(templatepath,"r") as f:
    text = f.read()
    text = text.replace("%JPI%",jpi)
    text = text.replace("%JPJ%",jpj)
    text = text.replace("%JPK%",jpk)
    text = text.replace("%RESOLUTION%",resolution)
    text = text.replace("%BATHYMETRY_FILE%",bathymetry_file)
    original_text = text[:]


# Try different cases
cases = []
# We try configurations that range from the maximum number of processes choosen to a 
# 130% of that value
for n in range(int(numprocs), int(numprocs*1.3)):
   # Decomposition of n=jpni*jpnj
   jpni,jpnj = decompose(n)
   # If the values are too much different, n is discarted.
   if 5 > float(jpni)/float(jpnj) > 0.2:
      # The jpni and jpnj values are modified in the template and written to a file
      text = original_text[:]
      text = text.replace("%JPNI%",str(jpni))
      text = text.replace("%JPNJ%",str(jpnj))
      with open("namelist","w") as out:
          out.write(text)
      # Temporary files needed for using subprocess call
      stdout=open(srcpath+"out","w")
      stderr=open(srcpath+"err","w")
      # Calling the binary with the modified namelist
      status = call([toolpath+"/bin/find_sea_procs.exe","2&> /dev/null "],stdout=stdout,stderr=stderr)
      # Checking for errors ( A cleaner way to do that would be great)
      if status:
         print "ERROR num %i" % status
      stdout.close()
      stderr.close()
      stderr=open(srcpath+"err","r")
      error_log = stderr.read()
      stderr.close()
      if error_log.find("shared libraries") > -1:
         print "ERROR:"
         print "Load proper environment"
         print 
         exit(1)
      

      # Open the file generated by the binary to parse the output
      with open("./fort.8") as inp:
       for line in inp:
          # Find sea processes
          if line.count("nombre de processeurs mer"):
            procs = int(line.split()[-1])
          # Find % of surface in decomposition
          if line.count("% sup"):
            coef = float(line.split()[-1])
      # Discarting the configurations with coef 0 (Erroneous)
      if procs <= int(numprocs):
         if coef == 0:
            coef = 10
         # Adding the result in the list of valid cases
         cases.append([jpni, jpnj ,n,procs, coef, procs/coef**2])

# Sorting the cases by computed coefficient
cases = sorted(cases,key=lambda x: x[-1])

# In case of printing simplified output, printing only the values of the best configuration without any description. This can be used in auto-nemo and auto-ecearth.
if simplified:
 case = list(reversed(cases))[0]
 jpi = case[0]
 jpj = case[1]
 SeaP = case[3]
 print SeaP,jpi,jpj

# In case of not using simplified output printing a brief description and the best L configurations found. L can be changed in command line argument -l 
else:
 print 
 print "Resolution ",resolution
 print "Max number of processors:", numprocs
 print 
 print "It will be better if you run with one of these configurations:"
 print "Bigger coefficient is better!"
 if len(cases) > list_length:
    cases = cases[-list_length:]
 for case in list(reversed(cases)):
  jpi = case[0]
  jpj = case[1]
  OriginalP = jpi*jpj
  SeaP = case[3]
  LandP = OriginalP - SeaP
  Coef = case[-1]
  print "jpij: %i jpi: %i jpj: %i  OriginalP: %i  LandP: %i SeaP: %i  Coeficient: %.2f" % ( SeaP, jpi, jpj, OriginalP, LandP, SeaP, Coef)
 
 

# Remove temporary files generated during the execution of this script.
files = os.listdir(".")
for f in files:
   if f.find(resolution+"-") ==0 and f.count("x") and f.count("_") and not f.count("."):
      os.remove(f)

os.remove("fort.8")
os.remove(srcpath+"err")
os.remove(srcpath+"out")
os.remove(toolpath+"namelist")
###################################################################
